"""
The "engine room" of django-yubin mailer.

Functions here actually handle the sending of messages.
"""

import logging

from django.core.mail import get_connection
from django.db import transaction

from . import models, settings


logger = logging.getLogger(__name__)


@transaction.atomic
def send_db_message(message_pk, log_message=None):
    """
    Sends a django_yubin.models.Message by its PK.
    """
    try:
        # Lock the message
        message = models.Message.objects.select_for_update().get(pk=message_pk)
    except Exception:
        msg = 'Could not fetch and lock the message from the database'
        logger.exception(msg, extra={'message_pk': message_pk})
        return False

    # Messages in STATUS_QUEUED can be sent to keep compatibility with previous yubin version.
    # In future versions that condition can be removed and only check `can_be_enqueued()`.
    if message.status != models.Message.STATUS_QUEUED:
        if not message.can_be_enqueued():
            msg = "Message can not be enqueued in it's current status."
            logger.warning(msg)
            message.add_log(msg)
            return False

    message.mark_as(models.Message.STATUS_QUEUED, log_message)
    message.mark_as(models.Message.STATUS_IN_PROCESS, "Trying to send the message.")

    recipients = message.recipients()
    if models.Blacklist.objects.filter(email__in=recipients).exists():
        msg = "Not sending due blacklisted email in: %s" % recipients
        logger.info(msg)
        message.mark_as(models.Message.STATUS_BLACKLISTED, msg)
        return False

    if settings.PAUSE_SEND:
        msg = "Sending is paused, discarding the email."
        logger.info(msg)
        message.mark_as(models.Message.STATUS_DISCARDED, msg)
        return False

    try:
        connection = get_connection(backend=settings.USE_BACKEND)
        connection.send_messages([message.get_email_message()])
        msg = "Message sent %s" % message
        logger.info(msg)
        message.mark_as(models.Message.STATUS_SENT, msg)
        return True
    except Exception as e:
        logger.exception("Message sending has failed", extra={'email_message': message})
        message.mark_as(models.Message.STATUS_FAILED, str(e))
        return False



import logging
import os
from abc import ABC, abstractmethod
from uuid import uuid4

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.utils.module_loading import import_string

from . import settings as yubin_settings
from .models import Message


logger = logging.getLogger(__name__)


class StorageBackendException(Exception):
    pass


class BaseStorageBackend(ABC):
    @classmethod
    @abstractmethod
    def get_message_data(cls, message): pass

    @classmethod
    @abstractmethod
    def set_message_data(cls, message, data): pass

    @classmethod
    @abstractmethod
    def delete_message_data(cls, message): pass

    @classmethod
    def admin_display_message_data(cls, model_admin, message):
        return f'''
            <textarea class="vLargeTextField" cols="40" rows="15" style="width: 99%;" disabled
            readonly>{message.message_data}</textarea>
        '''.strip()


class DatabaseStorageBackend(BaseStorageBackend):
    @classmethod
    def get_message_data(cls, message):
        return message._message_data

    @classmethod
    def set_message_data(cls, message, data):
        message._message_data = data

    @classmethod
    def delete_message_data(cls, message):
        pass


class FileStorageBackend(BaseStorageBackend):
    storage = default_storage

    @classmethod
    def get_message_data(cls, message):
        file = cls.storage.open(cls.get_path(message), 'rb')
        content = file.read().decode(settings.DEFAULT_CHARSET)
        file.close()
        return content

    @classmethod
    def set_message_data(cls, message, data):
        path = cls.get_path(message)
        new_path = cls.storage.save(path, ContentFile(data.encode(settings.DEFAULT_CHARSET)))
        if message._message_data:
            cls.storage.delete(message._message_data)
        message._message_data = new_path

    @staticmethod
    def get_path(message):
        return message._message_data or \
            os.path.join(yubin_settings.MAILER_FILE_STORAGE_DIR, f"{str(uuid4())}.msg")

    @classmethod
    def delete_message_data(cls, message):
        cls.storage.delete(cls.get_path(message))

    @classmethod
    def admin_display_message_data(cls, model_admin, message):
        return f'''
            <div>
                <a href="{cls.storage.url(message._message_data)}">
                    {message._message_data}
                </a>
            </div>
            <br>
            {BaseStorageBackend.admin_display_message_data(model_admin, message)}
        '''.strip()


def db2file():
    """
    Migrate emails from DatabaseStorageBackend to FileStorageBackend.
    """
    backend = import_string(yubin_settings.MAILER_STORAGE_BACKEND)
    if backend != FileStorageBackend:
        raise StorageBackendException(
            f'settings.MAILER_STORAGE_BACKEND should be {FileStorageBackend} instead of {backend}')

    messages = Message.objects \
        .filter(storage='django_yubin.storage_backends.DatabaseStorageBackend') \
        .only('pk', '_message_data', 'storage')
    for message in messages:
        db_message_data = DatabaseStorageBackend.get_message_data(message)
        DatabaseStorageBackend.set_message_data(message, data='')
        message.storage = yubin_settings.MAILER_STORAGE_BACKEND
        message.message_data = db_message_data
        message.save()
        logger.info(f'Message {message.pk} migrated. Saved in {message._message_data}')


def file2db(delete=False):
    """
    Migrate emails from FileStorageBackend to DatabaseStorageBackend optionally deleting the files
    from the file storage.
    """
    backend = import_string(yubin_settings.MAILER_STORAGE_BACKEND)
    if backend != DatabaseStorageBackend:
        raise StorageBackendException(
            f'settings.MAILER_STORAGE_BACKEND should be {DatabaseStorageBackend} instead of {backend}')

    messages = Message.objects \
        .filter(storage='django_yubin.storage_backends.FileStorageBackend') \
        .only('pk', '_message_data', 'storage')
    for message in messages:
        file_message_path = FileStorageBackend.get_path(message)
        file_message_data = FileStorageBackend.get_message_data(message)
        message.storage = yubin_settings.MAILER_STORAGE_BACKEND
        message.message_data = file_message_data
        message.save()
        logger.info(f'Message {message.pk} migrated')

        if delete:
            FileStorageBackend.storage.delete(file_message_path)
            logger.info(f'File {file_message_path} deleted from FileStorageBackend')



import datetime
import logging
import email
from email import policy
from email import encoders as Encoders
from email.mime.base import MIMEBase
from functools import partial

from django.core.exceptions import FieldError
from django.core.mail.message import (
        ADDRESS_HEADERS,
        EmailMessage,
        EmailMultiAlternatives,
    )
from django.db import models, transaction
from django.db.models import F
from django.utils.module_loading import import_string
from django.utils.text import Truncator
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from mailparser import MailParser

from . import mailparser_utils, tasks


logger = logging.getLogger(__name__)


PARSED_HEADERS_TO_IGNORE = ADDRESS_HEADERS.union({"content-type", "subject", "mime-version"})


class MessageQuerySet(models.QuerySet):
    def retryable(self, max_retries=0):
        qs = self.filter(status__gte=self.model.STATUS_FAILED)
        if max_retries > 0:
            qs = qs.filter(enqueued_count__lt=max_retries)
        return qs


class MessageManager(models.Manager):
    def get_queryset(self):
        return MessageQuerySet(self.model, using=self._db)

    def retryable(self, max_retries=0):
        return self.get_queryset().retryable(max_retries=max_retries)


class Message(models.Model):
    """
    An email message.
    """
    STATUS_CREATED = 0
    STATUS_QUEUED = 1
    STATUS_IN_PROCESS = 2
    STATUS_SENT = 3
    STATUS_FAILED = 4
    STATUS_BLACKLISTED = 5
    STATUS_DISCARDED = 6
    STATUS_CHOICES = (
        (STATUS_CREATED, _('Created')),
        (STATUS_QUEUED, _('Queued')),
        (STATUS_IN_PROCESS, _('In process')),
        (STATUS_SENT, _('Sent')),
        (STATUS_FAILED, _('Failed')),
        (STATUS_BLACKLISTED, _('Blacklisted')),
        (STATUS_DISCARDED, _('Discarded')),
    )

    to_address = models.TextField(_('to addresses'))
    cc_address = models.TextField(_('cc addresses'), blank=True, default="")
    bcc_address = models.TextField(_('bcc addresses'), blank=True, default="")

    # These fields are to easy access, filtering and searching without parsing the email.
    from_address = models.CharField(_('from address'), max_length=200)
    subject = models.CharField(_('subject'), max_length=255)

    # This field is for internal use in storage backends. They can use it to save the email
    # like the DatabaseSorageBackend, the file path like the FileStorageBackend, etc.
    # Other users must access this data through the ``message_data`` property.
    _message_data = models.TextField(_('message data'), db_column='message_data')

    @property
    def message_data(self):
        storage_backend = import_string(self.storage)
        return storage_backend.get_message_data(self)

    @message_data.setter
    def message_data(self, data):
        storage_backend = import_string(self.storage)
        storage_backend.set_message_data(self, data)

    storage = models.CharField(_('storage backend'), max_length=200, blank=False,
                               default="django_yubin.storage_backends.DatabaseStorageBackend")

    date_created = models.DateTimeField(_('date created'), auto_now_add=True)
    date_sent = models.DateTimeField(_('date sent'), null=True, blank=True)
    sent_count = models.PositiveSmallIntegerField(_('sent count'), default=0,
                                                  help_text=_('Times the message has been sent'))
    date_enqueued = models.DateTimeField(_('date enqueued'), null=True, blank=True)
    enqueued_count = models.PositiveSmallIntegerField(_('enqueued count'), default=0,
                                                      help_text=_('Times the message has been enqueued'))
    status = models.PositiveSmallIntegerField(choices=STATUS_CHOICES, default=STATUS_CREATED)

    objects = MessageManager()

    class Meta:
        ordering = ('date_created',)
        verbose_name = _('message')
        verbose_name_plural = _('messages')

    def __init__(self, *args, **kwargs):
        if '_message_data' in kwargs:
            raise FieldError("_message_data can not be used for creating instances, use message_data.")
        return super().__init__(*args, **kwargs)

    def __str__(self):
        recipients = self.to_address
        if self.cc_address:
            recipients += ', %s' % self.cc_address
        if self.bcc_address:
            recipients += ', %s' % self.bcc_address
        return '%s: %s' % (recipients, self.subject)

    def to(self):
        return [email.strip() for email in self.to_address.split(",") if email.strip()]

    def cc(self):
        return [email.strip() for email in self.cc_address.split(",") if email.strip()]

    def bcc(self):
        return [email.strip() for email in self.bcc_address.split(",") if email.strip()]

    def recipients(self):
        return self.to() + self.cc() + self.bcc()

    def get_message_parser(self):
        message = email.message_from_string(self.message_data, policy=policy.default)
        return MailParser(message)

    def get_email_message(self):
        """
        Returns EmailMultiAlternatives or EmailMessage depending on whether the email is multipart or not.

        Note that this should reconstruct all the kwargs of EmailMessage/EmailMultiAlternatives
        to not lose information. See Github issue #66 for more information.

        XXX: perhaps an inspect.signature combined with **kwargs approach could validate
        against unexpected django changes?
        """
        msg = self.get_message_parser()
        to = self.to() or mailparser_utils.get_addresses(msg.to)
        cc = self.cc() or mailparser_utils.get_addresses(msg.cc)
        bcc = self.bcc()

        # Process headers, but ignore address headers - these are processed explicitly.
        headers = {
            header: value
            for header, value in msg.headers.items()
            if header.lower() not in PARSED_HEADERS_TO_IGNORE
        }

        Email = EmailMultiAlternatives if msg.text_html else EmailMessage
        email = Email(
            subject=msg.subject,
            body='\n'.join(msg.text_plain),
            from_email=mailparser_utils.get_address(msg.from_),
            to=to,
            bcc=bcc,
            headers=headers,
            cc=cc,
            reply_to=mailparser_utils.get_addresses(msg.reply_to),
        )

        # set the multipart subtype
        content_type = msg.headers["Content-Type"].split(";", 1)[0]  # discard boundary
        main_type, subtype = content_type.split("/", 1)
        if main_type == "multipart":
            email.mixed_subtype = subtype

        # NOTE - mailparser only supports text and HTML, any other content types are
        # considered not_managed.
        if msg.text_html:
            email.attach_alternative('<br>'.join(msg.text_html), mimetype='text/html')

        # attachment is a dict with fixed keys:
        # filename, payload, binary, mail_content_type, content-id, content-disposition,
        # charset and content_transfer_encoding
        #
        # This performs generic handling of attachments, respecting the original various
        # ways the attachment can be used.
        for attachment in msg.attachments:
            basetype, subtype = attachment["mail_content_type"].split("/", 1)
            binary = attachment["binary"]
            content = attachment['payload']
            transfer_encoding = attachment["content_transfer_encoding"]

            mime_attachment = MIMEBase(basetype, subtype)
            mime_attachment.set_payload(content)
            if not binary:
                Encoders.encode_base64(mime_attachment)
            else:
                mime_attachment.add_header("Content-Transfer-Encoding", transfer_encoding)
            for header in ("Content-ID", "Content-Disposition"):
                value = attachment[header.lower()]
                if value:
                    mime_attachment.add_header(header, value)
            email.attach(mime_attachment)

        return email

    def add_log(self, log_message):
        Log.objects.create(message=self, action=self.status, log_message=log_message)

    def mark_as(self, status, log_message=None):
        should_refresh_from_db = False

        self.status = status
        if status is self.STATUS_SENT:
            self.date_sent = now()
            self.sent_count = F('sent_count') + 1
            should_refresh_from_db = True
        elif status is self.STATUS_QUEUED:
            self.date_enqueued = now()
            self.enqueued_count = F('enqueued_count') + 1
            should_refresh_from_db = True
        self.save()

        if should_refresh_from_db:
            self.refresh_from_db()

        if log_message is not None:
            self.add_log(log_message)

    def can_be_enqueued(self):
        return self.status in (
            self.STATUS_CREATED,
            self.STATUS_FAILED,
            self.STATUS_BLACKLISTED,
            self.STATUS_DISCARDED
        )

    def enqueue(self, log_message=None):
        """
        Sends the task to enqueue the message on commit.
        """
        if not self.can_be_enqueued():
            self.add_log("Message can not be enqueued in it's current status")
            return False

        transaction.on_commit(partial(tasks.send_email.delay, message_pk=self.pk, log_message=log_message))
        return True

    @classmethod
    def retry_messages(cls, max_retries=3):
        enqueued = 0
        messages = cls.objects.retryable(max_retries)
        for message in messages:
            enqueued += message.enqueue('Retry sending the email.')
        failed = len(messages) - enqueued
        return enqueued, failed

    @classmethod
    def delete_old(cls, days=90):
        """
        Deletes mails created before `days` days.

        Returns the deletion data from Django and the cuttoff date.
        """
        cutoff_date = now() - datetime.timedelta(days)
        deleted = cls.objects.filter(date_created__lt=cutoff_date).delete()
        return deleted, cutoff_date


class Blacklist(models.Model):
    """
    A blacklisted email address.

    Messages attempted to be sent to e-mail addresses which appear on this
    blacklist will be skipped entirely.
    """
    email = models.EmailField(_('email'), max_length=200)
    date_added = models.DateTimeField(_('date added'), default=now)

    class Meta:
        ordering = ('-date_added',)
        verbose_name = _('blacklisted email')
        verbose_name_plural = _('blacklisted emails')

    def __str__(self):
        return self.email


class Log(models.Model):
    """
    A log used to record the activity of a queued message.
    """
    message = models.ForeignKey(Message, verbose_name=_('message'), editable=False, on_delete=models.CASCADE)
    action = models.PositiveSmallIntegerField(_('action'), choices=Message.STATUS_CHOICES,
                                              default=Message.STATUS_CREATED)
    date = models.DateTimeField(_('date'), auto_now_add=True)
    log_message = models.TextField(_('log'), blank=True)

    class Meta:
        ordering = ('-date',)
        verbose_name = _('log')
        verbose_name_plural = _('logs')

    def __str__(self):
        return Truncator(self.log_message).chars(50)



from django.conf import settings


# Provide a way of temporarily pausing the sending of mail.
PAUSE_SEND = getattr(settings, "MAILER_PAUSE_SEND", False)

# Real backend to send emails.
USE_BACKEND = getattr(settings, 'MAILER_USE_BACKEND', 'django.core.mail.backends.smtp.EmailBackend')

# When MAILER_TEST_MODE is True, recipient addresses of all messages are replaced with
# the email addresses set in MAILER_TEST_EMAIL before being sent.
MAILER_TEST_MODE = getattr(settings, "MAILER_TEST_MODE", False)
MAILER_TEST_EMAIL = getattr(settings, "MAILER_TEST_EMAIL", '')

# If there are emails created, enqueued or in progress for more than x minutes, the HealthCheck
# view will show an error.
MAILER_HC_QUEUED_LIMIT_OLD = getattr(settings, "MAILER_HC_QUEUED_LIMIT_OLD", 30)

# Storage to save full emails.
MAILER_STORAGE_BACKEND = getattr(
    settings,
    "MAILER_STORAGE_BACKEND",
    "django_yubin.storage_backends.DatabaseStorageBackend",
)

# Delete storage data when deleting messages from the database.
MAILER_STORAGE_DELETE = getattr(settings, "MAILER_STORAGE_DELETE", True)

# Subdirectory to save emails when using the FileStorageBackend.
MAILER_FILE_STORAGE_DIR = getattr(settings, "MAILER_FILE_STORAGE_DIR", 'yubin')



from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.module_loading import import_string

from . import settings as yubin_settings


@receiver(post_delete, sender='django_yubin.Message', dispatch_uid='django_yubin_delete_storage')
def delete_message_storage_callback(sender, **kwargs):
    """
    Deleting storage objects in the post_delete signal assures that they will be deleted always:
    from a single object, from a queryset or from the admin.

    https://docs.djangoproject.com/en/3.2/topics/db/models/#overriding-predefined-model-methods
    """
    if yubin_settings.MAILER_STORAGE_DELETE:
        message = kwargs['instance']
        storage_backend = import_string(message.storage)
        storage_backend.delete_message_data(message)



from django.core.management.base import BaseCommand, CommandError

from ...storage_backends import db2file, StorageBackendException


class Command(BaseCommand):
    help = 'Migrate emails from DatabaseStorageBackend storage to FileStorageBackend.'

    def handle(self, *args, **options):
        try:
            db2file()
        except StorageBackendException as e:
            raise CommandError(e)



from django.apps import AppConfig


class YubinConfig(AppConfig):
    name = 'django_yubin'
    verbose_name = 'Django Yubin'
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):
        from . import signals  # noqa



# Generated by Django 4.2 on 2023-05-03 07:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_yubin", "0010_message_storage")]

    operations = [
        migrations.AddField(
            model_name="message",
            name="bcc_address",
            field=models.TextField(
                blank=True, default="", verbose_name="bcc addresses"
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="cc_address",
            field=models.TextField(blank=True, default="", verbose_name="cc addresses"),
        ),
        migrations.AlterField(
            model_name="message",
            name="to_address",
            field=models.TextField(verbose_name="to addresses"),
        ),
    ]



# Generated by Django 4.2.4 on 2024-06-21 08:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_yubin', '0011_message_bcc_address_message_cc_address_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='blacklist',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='log',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='message',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
